C     ******************************************************************
C
C              ROBOTFINDSKITTEN (RFK) IN FORTRAN 77 FOR MS-DOS
C                        AND TENBERRY SOFTWARE DOS/4G
C
C                     Copyright (C) Philipp Engel, 2021
C                              Version 1.1.200
C
C     ******************************************************************
C
C     This program is a port of "robotfindskitten" in FORTRAN 77 for
C     32-bit DOS. The simulation is compliant to RFC-RFK-00:
C
C         "Standard for robotfindskitten Zen Simulation Environments"
C
C     For the original implementation, please see:
C
C         http://robotfindskitten.org/
C
C     Run Open Watcom FORTRAN 77 to compile the source code:
C
C         > wfl386 src\rfk.for /l=dos4g
C         > rfk.exe
C
C     ******************************************************************
C
*$pragma aux delay "*_" parm (value)
C
C     ******************************************************************
      PROGRAM RFK
      INTEGER  BIOSDT, GETITE
      LOGICAL  AGAIN
      EXTERNAL ANIMAT, CLS, CURSOR, INSTRU, KEYPRS, LDFILE, SRAND
      EXTERNAL FREEXY, INITFD
      EXTERNAL OUTCHR, OUTITE, OUTSTR

      CHARACTER*72 FNAME
      INTEGER      MAJOR, MINOR, NNKI
      INTEGER      OFFX, OFFY
      INTEGER      WIDTH, HEIGHT
      PARAMETER (FNAME='nki\fortran.nki')
      PARAMETER (MAJOR=1, MINOR=1, NNKI=200)
      PARAMETER (OFFX=-1, OFFY=2)
      PARAMETER (WIDTH=80, HEIGHT=22)

      CHARACTER*72 DESC(NNKI)
      CHARACTER*1  KEYCHR, ROBOCH
      INTEGER      ROBOX, ROBOY
      INTEGER      ROBOCL
      INTEGER      FIELD(WIDTH, HEIGHT)
      INTEGER      ICH, ICL, IDI
      INTEGER      CODE, I, SEED, STAT
      INTEGER      DX, DY, NX, NY
      LOGICAL      CTRL, DONE, KITTEN
C
C     Initialise PRNG.
C
      SEED = BIOSDT()
      CALL SRAND(SEED)
C
C     Load NKI file.
C
      CALL LDFILE(NNKI, DESC, FNAME, STAT)
      IF (STAT .NE. 0) STOP 'Error: Loading NKI file failed'
C
C     Output version and introduction.
C
      CALL CLS()
      CALL CURSOR(.FALSE.)
      CALL OUTVER(MAJOR, MINOR, NNKI, .TRUE.)
      CALL INSTRU()
      CALL KEYPRS(KEYCHR, CODE, CTRL)
C
C     Main loop.
C
      ROBOCH = CHAR(35)
      ROBOCL = 7
      DONE   = .FALSE.

      DO WHILE (.NOT. DONE)
        CALL CLS()
        CALL OUTVER(MAJOR, MINOR, NNKI, .FALSE.)

        DO I = 0, WIDTH - 1
          CALL OUTCHR(I, 2, '-', 7, 0)
        END DO
C
C     Place NKIs and kitten on field.
C
        CALL INITFD(WIDTH, HEIGHT, FIELD, 24, NNKI)
C
C     Place player somewhere on the field.
C
        CALL FREEXY(WIDTH, HEIGHT, FIELD, ROBOX, ROBOY)
C
C     Output items and robot.
C
        CALL OUTITE(WIDTH, HEIGHT, FIELD, OFFX, OFFY)
        CALL OUTCHR(ROBOX + OFFX, ROBOY + OFFY, ROBOCH, ROBOCL, 0)

        DO
          DX = 0
          DY = 0

          CALL KEYPRS(KEYCHR, CODE, CTRL)

          SELECT CASE (CODE)
            CASE (16)
C             Quit (Q):
              DONE = .TRUE.
              EXIT
            CASE (46)
C             Quit (CTRL + C):
              IF (CTRL) THEN
                DONE = .TRUE.
                EXIT
              END IF
            CASE (72)
C             Move up:
              DY = -1
            CASE (75)
C             Move left:
              DX = -1
            CASE (77)
C             Move right:
              DX = 1
            CASE (80)
C             Move down:
              DY = 1
          END SELECT

          IF (DX .EQ. 0 .AND. DY .EQ. 0) CYCLE

          NX = ROBOX + DX
          NY = ROBOY + DY

          IF (NX .LT. 1 .OR. NX .GT. WIDTH .OR.
     &        NY .LT. 1 .OR. NY .GT. HEIGHT) CYCLE

          N = GETITE(FIELD(NX, NY), ICH, ICL, IDI, KITTEN)

          IF (N .GT. 0) THEN
            IF (KITTEN) THEN
              CALL ANIMAT(ICH, ICL)
              CALL OUTSTR(0, 1, 'You found kitten! Way to go, robot!',
     &                    7, 0)
              CALL DELAY(1000)
              EXIT
            ELSE
              CALL OUTSTR(0, 1, DESC(IDI), 7, 0)
            END IF
          ELSE
            DO I = 0, 71
              CALL OUTCHR(I, 1, ' ', 7, 0)
            END DO
            CALL OUTCHR(ROBOX + OFFX, ROBOY + OFFY, ' ', 0, 0)
            ROBOX = NX
            ROBOY = NY
            CALL OUTCHR(ROBOX + OFFX, ROBOY + OFFY, ROBOCH, ROBOCL, 0)
          END IF
        END DO
        IF (.NOT. DONE) THEN
          IF (.NOT. AGAIN()) DONE = .TRUE.
        END IF
      END DO

      CALL CLS()
      END
C
C     ******************************************************************
C
      FUNCTION BIOSDT()
C
C     Returns the number of clock ticks since midnight.
C
C
      INTEGER BIOSDT
      INCLUDE 'DOS.FI'
      INTEGER INT
      PARAMETER (INT='1A'X)

      AH = 0
      CALL FINTR(INT, REGS)
      BIOSDT = IOR(ISHFT(CX, 16), DX)
      END
C
C     ******************************************************************
C
      FUNCTION AGAIN()
C
C     Returns whether or not the player wants another game.
C
      EXTERNAL KEYPRS, OUTSTR
      LOGICAL AGAIN
      CHARACTER*1 CHR
      INTEGER     CODE
      LOGICAL     CTRL

      CALL OUTSTR(0, 0, 'Do you want to play again? (Y/N)', 15, 0)

      DO
        CALL KEYPRS(CHR, CODE, CTRL)
        IF (CHR .EQ. 'Y' .OR. CHR .EQ. 'y') THEN
          AGAIN = .TRUE.
          EXIT
        ELSE IF (CHR .EQ. 'N' .OR. CHR .EQ. 'n') THEN
          AGAIN = .FALSE.
          EXIT
        END IF
      END DO
      END
C
C     ******************************************************************
C
      FUNCTION GETITE(I, ICH, ICL, IDI, KITTEN)
C
C     Returns given item and its character, colour, description index,
C     and kitten status.
C
C     Arguments:
C
C         I         -   Item.
C         ICH       -   Item character as integer (0 ... 255).
C         ICL       -   Item colour (0 ... 14)
C         IDI       -   Item description index.
C
      INTEGER GETITE
      INTEGER I, ICH, ICL, IDI
      LOGICAL KITTEN

      ICH = 33
      ICL = 1
      IDI = 1
      KITTEN = .FALSE.
      GETITE = I

      IF (I .GT. 0) THEN
        ICL = MOD(I / 10, 100)
        ICH = MOD(I / 1000, 1000)
        IF (MOD(I, 10) .EQ. 1) THEN
          KITTEN = .TRUE.
        ELSE
          IDI = INT(I / 10**6)
        END IF
      END IF
      END
C
C     ******************************************************************
C
      SUBROUTINE ANIMAT(ICH, ICL)
C
C     Outputs the final animation.
C
C     Arguments:
C
C         ICH       -   Character (as integer).
C         ICL       -   Foreground colour (0 ... 15).
C
      EXTERNAL DELAY, OUTCHR, OUTSTR
      INTEGER ICH, ICL
      INTEGER I

      DO I = 1, 4
        CALL OUTSTR(72, 1, '        ', 0, 0)
        CALL OUTCHR(71 + I, 1, '#', 7, 0)
        CALL OUTCHR(80 - I, 1, CHAR(ICH), ICL, 0)
        CALL DELAY(1000)
      END DO
      CALL OUTSTR(75, 0, CHAR(3) // CHAR(3), 12, 0)
      END
C
C     ******************************************************************
C
      SUBROUTINE CLS()
C
C     Clears the screen by calling BIOS interrupt 10h.
C
      INCLUDE 'DOS.FI'
      INTEGER INT
      PARAMETER (INT='10'X)

      AH = '06'X
      AL = '00'X
      BH = '07'X
      DX = '184F'X
      CALL FINTR(INT, REGS)
      END
C
C     ******************************************************************
C
      SUBROUTINE CURSOR(ENABLE)
C
C     Shows or hides the text cursor by calling BIOS interrupt 10h.
C
C     Arguments:
C
C         ENABLE    -   Flag to enable cursor.
C
      INCLUDE 'DOS.FI'
      LOGICAL ENABLE
      INTEGER INT
      PARAMETER (INT='10'X)

      AH = '01'X
      IF (ENABLE) THEN
        CX = '0607'X
      ELSE
        CX = '2607'X
      END IF
      CALL FINTR(INT, REGS)
      END
C
C     ******************************************************************
C
      SUBROUTINE FREEXY(WIDTH, HEIGHT, FIELD, X, Y)
C
C     Returns free field at position X, Y.
C
C     Arguments:
C
C         WIDTH     -   Columns in FIELD.
C         HEIGHT    -   Rows in FIELD.
C         FIELD     -   Array of items.
C         X         -   Column in FIELD or -1.
C         Y         -   Row in FIELD or -1.
C
      REAL RAND
      INTEGER WIDTH, HEIGHT, FIELD(WIDTH, HEIGHT), X, Y
      INTEGER I

      I = 1
      DO
        IF (I .GT. 1000) THEN
          X = -1
          Y = -1
          EXIT
        END IF
        X = 1 + NINT(RAND() * (WIDTH - 1))
        Y = 1 + NINT(RAND() * (HEIGHT - 1))
        IF (FIELD(X, Y) .EQ. 0) EXIT
        I = I + 1
      END DO
      END
C
C     ******************************************************************
C
      SUBROUTINE INITFD(WIDTH, HEIGHT, FIELD, N, NNKI)
C
C     Places N NKIs and a single kitten on the field.
C
C     Arguments:
C
C         WIDTH     -   Columns in FIELD.
C         HEIGHT    -   Rows in FIELD.
C         FIELD     -   Array of items.
C         N         -   Number of items to place.
C         NNKI      -   Number of non-kitten items available.
C
      REAL     RAND
      EXTERNAL FREEXY
      INTEGER WIDTH, HEIGHT, FIELD(WIDTH, HEIGHT), N, NNKI
      INTEGER CL, CH, ID, KI
      INTEGER X, Y
      INTEGER I, J

      DO J = 1, HEIGHT
        DO I = 1, WIDTH
          FIELD(I, J) = 0
        END DO
      END DO

      DO I = 1, N
        IF (I .GT. N) EXIT
        CALL FREEXY(WIDTH, HEIGHT, FIELD, X, Y)
        CL = 1 + NINT(RAND() * 14)
        CH = 33 + NINT(RAND() * 92)
        ID = 1
        KI = 0

        IF (I .EQ. N) THEN
          KI = 1
        ELSE
          ID = 1 + NINT(RAND() * (NNKI - 1))
        END IF

        FIELD(X, Y) = KI + (CL * 10) + (1000 * CH) + (10**6 * ID)
      END DO
      END
C
C     ******************************************************************
C
      SUBROUTINE INSTRU()
C
C     Displays instructions on screen.
C
      INTEGER*1 BG, FG
      PARAMETER (BG=0, FG='F'X)

      CALL OUTSTR(0, 4, 'In this game, you are robot (#). Your job ' //
     &  'is to find kitten. This task', FG, BG)
      CALL OUTSTR(0, 5, 'is complicated by the existence of various ' //
     &  'things which are not kitten.', FG, BG)
      CALL OUTSTR(0, 6, 'Robot must touch items to determine if ' //
     &  'they are kitten or not. The game', FG, BG)
      CALL OUTSTR(0, 7, 'ends when robotfindskitten. Alternatively, ' //
     &  'you may end the game by', FG, BG)
      CALL OUTSTR(0, 8, 'hitting the q key or a good old-fashioned ' //
     &  'Ctrl-C.', FG, BG)
      CALL OUTSTR(0, 10, 'See the documentation for more information.',
     &  FG, BG)
      CALL OUTSTR(0, 12, 'Press any key to start.', 7, BG)
      END
C
C     ******************************************************************
C
      SUBROUTINE KEYPRS(CHR, CODE, CTRL)
C
C     Returns character and scan code of the key pressed (blocking).
C
C     Arguments:
C
C         CHR       -   The character of the key pressed.
C         CODE      -   The scan code.
C         CTRL      -   Flag for CTRL key pressed.
C
      INCLUDE 'DOS.FI'
      INTEGER INT
      PARAMETER (INT='16'X)
      CHARACTER*1 CHR
      INTEGER     CODE
      LOGICAL     CTRL

      AH = '00'X
      CALL FINTR(INT, REGS)
      CHR  = CHAR(AL)
      CODE = AH
      CTRL = .FALSE.

      AH = '02'X
      CALL FINTR(INT, REGS)
      IF (IAND(ISHFT('1'X, 2), AL)) CTRL = .TRUE.
      END
C
C     ******************************************************************
C
      SUBROUTINE LDFILE(N, DESCR, FNAME, STAT)
C
C     Loads N non-kitten item descriptions from file into array DESCR.
C
C     Arguments:
C
C         N         -   Number of NKI.
C         DESCR     -   Array of size N to store NKI descriptions.
C         FNAME     -   Path to the NKI file.
C         STAT      -   I/O status.
C
      INTEGER       N, STAT
      CHARACTER*72  DESCR(N)
      CHARACTER*(*) FNAME
      INTEGER       I

      OPEN (UNIT=10, FILE=FNAME, IOSTAT=STAT, STATUS='OLD')
      IF (STAT .NE. 0) RETURN
      DO I = 1, N
        READ (UNIT=10, FMT=100, IOSTAT=STAT) DESCR(I)
  100   FORMAT (A)
        IF (STAT .NE. 0) EXIT
      END DO
      CLOSE (UNIT=10)
      END
C
C     ******************************************************************
C
      SUBROUTINE OUTCHR(X, Y, CHR, FGCOL, BGCOL)
C
C     Outputs a singe character at position X, Y with given foreground
C     and background color.
C
C     Arguments:
C
C         X         -   Column (0 ... 79).
C         Y         -   Row (0 ... 24).
C         CHR       -   Character to output (33 ... 126).
C         FGCOL     -   Foreground colour (0 ... 15).
C         BGCOL     -   Background colour (0 ... 7).
C
      INTEGER COLS, ROWS, SCRSZE
      PARAMETER (COLS=80, ROWS=25)
      PARAMETER (SCRSZE=COLS * ROWS)
      INTEGER     X, Y
      CHARACTER*1 CHR
      INTEGER*1   FGCOL, BGCOL
      CHARACTER*1 SCREEN(:, :)
      INTEGER*1   ATTR
      INTEGER     I
      SAVE        SCREEN

      IF (.NOT. ALLOCATED(SCREEN))
     &  ALLOCATE (SCREEN(0:1, 0:SCRSZE - 1), LOCATION='B8000'X)

      I = Y * COLS + X
      IF (I .LT. 0 .OR. I .GT. SCRSZE - 1) RETURN
      ATTR = IOR(ISHFT(BGCOL, 4), IAND(FGCOL, '0F'X))

      SCREEN(0, I) = CHR
      SCREEN(1, I) = CHAR(ATTR)
      END
C
C     ******************************************************************
C
      SUBROUTINE OUTITE(WIDTH, HEIGHT, FIELD, OFFX, OFFY)
C
C     Outputs all items in FIELD to screen, with offset OFFX, OFFY.
C
C     Arguments:
C
C         WIDTH     -   Columns in FIELD.
C         HEIGHT    -   Rows in FIELD.
C         FIELD     -   Array of items.
C         OFFX      -   Column offset.
C         OFFY      -   Row offset.
C
      INTEGER  GETITE
      EXTERNAL OUTCHR
      INTEGER WIDTH, HEIGHT, FIELD(WIDTH, HEIGHT), OFFX, OFFY
      INTEGER CH, CL, ID, N, X, Y
      LOGICAL KITTEN

      DO Y = 1, HEIGHT
        DO X = 1, WIDTH
          N = GETITE(FIELD(X, Y), CH, CL, ID, KITTEN)
          IF (N .GT. 0) THEN
            CALL OUTCHR(X + OFFX, Y + OFFY, CHAR(CH), CL, 0)
          ELSE
            CALL OUTCHR(X + OFFX, Y + OFFY, ' ', 0, 0)
          END IF
        END DO
      END DO
      END
C
C     ******************************************************************
C
      SUBROUTINE OUTSTR(X, Y, STR, FGCOL, BGCOL)
C
C     Outputs string to screen at given position X, Y, in foreground
C     and background colour.
C
C     Arguments:
C
C         X         -   Column (0 .. 79)
C         Y         -   Row (0 .. 24)
C         STR       -   String to output.
C         FGCOL     -   Foreground colour (0 ... 15)
C         BGCOL     -   Background colour (0 ... 7)
C
      EXTERNAL OUTCHR
      INTEGER       X, Y
      CHARACTER*(*) STR
      INTEGER*1     FGCOL, BGCOL
      INTEGER       I

      DO I = 1, LEN(STR)
        CALL OUTCHR(X + I - 1, Y, STR(I:I), FGCOL, BGCOL)
      END DO
      END
C
C     ******************************************************************
C
      SUBROUTINE OUTVER(MAJOR, MINOR, NNKI, COPYR)
C
C     Outputs the program version to screen.
C
C     Arguments:
C
C         MAJOR     -   Major version.
C         MINOR     -   Minor version.
C         NNKI      -   Number of unique NKI descriptions.
C         COPYR     -   Output copyright or not.
C
      INTEGER  MAJOR, MINOR, NNKI
      LOGICAL  COPYR
      EXTERNAL OUTSTR
      CHARACTER*7  V

      WRITE (V, 100) MAJOR, MINOR, NNKI
  100 FORMAT (I1, '.', I1, '.', I3)
      CALL OUTSTR(0, 0, 'robotfindskitten, v. ' // V, 7, 0)
      IF (COPYR) THEN
        CALL OUTSTR(0, 1, 'FORTRAN 77 version by Philipp Engel (C) ' //
     &  '2020', 7, 0)
        CALL OUTSTR(0, 2, 'Original implementation by the ' //
     &  'illustrious Leonard Richardson (C) 1997, 2000', 7, 0)
      END IF
      END
C
C     ******************************************************************
C
      SUBROUTINE SRAND(ISEED)
C
C     This subroutine sets the integer seed to be used with the
C     companion RAND function to the value of ISEED. A flag is set to
C     indicate that the sequence of pseudo-random numbers for the
C     specified seed should start from the beginning.
C
      COMMON /SEED/ JSEED, IFRST
      JSEED = ISEED
      IFRST = 0
      END
C
C     ******************************************************************
C
      REAL FUNCTION RAND()
C
C     This function returns a pseudo-random number for each invocation.
C     It is a FORTRAN 77 adaptation of the "Integer Version 2" minimal
C     standard number generator whose Pascal code appears in the
C     article:
C
C         Park, Steven K. and Miller, Keith W., "Random Number
C         Generators: Good Ones are Hard to Find", Communications of
C         the ACM, October, 1988.
C
C     Original FORTRAN 77 implementation by Dick Valent and Fred Clare.
C
      PARAMETER (MPLIER=16807, MODLUS=2147483647, MOBYMP=127773,
     &           MOMDMP=2836)
      COMMON  /SEED/ JSEED, IFRST
      INTEGER HVLUE, LVLUE, TESTV, NEXTN
      SAVE    NEXTN

      IF (IFRST .EQ. 0) THEN
          NEXTN = JSEED
          IFRST = 1
      ENDIF

      HVLUE = NEXTN / MOBYMP
      LVLUE = MOD(NEXTN, MOBYMP)
      TESTV = MPLIER * LVLUE - MOMDMP * HVLUE

      IF (TESTV .GT. 0) THEN
          NEXTN = TESTV
      ELSE
          NEXTN = TESTV + MODLUS
      END IF

      RAND = REAL(NEXTN) / REAL(MODLUS)
      END
C
C     ******************************************************************
C
      BLOCKDATA RANDBD
      COMMON /SEED/ JSEED, IFRST
      DATA JSEED, IFRST /123456789, 0/
      END
C
C     ******************************************************************
